@@ -2,9 +2,11 @@ from sqlvalidator.grammar.sql import (
     Alias,
     AnalyticsClause,
     ArithmaticOperator,
+    BitwiseOperation,
     Boolean,
     BooleanCondition,
     Case,
+    CastFunctionCall,
     Column,
     CombinedQueries,
     Condition,
@@ -28,6 +30,7 @@ from sqlvalidator.grammar.sql import (
     String,
     Table,
     Type,
+    Unnest,
     UsingClause,
     WhereClause,
     WindowFrameClause,
@@ -36,6 +39,7 @@ from sqlvalidator.grammar.tokeniser import (
     get_tokens_until_closing_parenthesis,
     get_tokens_until_not_in,
     get_tokens_until_one_of,
+    lower,
 )
 
 
@@ -47,7 +51,7 @@ class SQLStatementParser:
     @staticmethod
     def parse(tokens):
         next_token = next(tokens)
-        if next_token == "select":
+        if lower(next_token) == "select":
             return SelectStatementParser.parse(tokens)
         raise ParsingError
 
@@ -62,12 +66,12 @@ class SelectStatementParser:
 
         select_all = select_distinct = False
         select_distinct_on = None
-        if next_token == "all":
+        if lower(next_token) == "all":
             select_all = True
-        elif next_token == "distinct":
+        elif lower(next_token) == "distinct":
             select_distinct = True
             next_token = next(tokens)
-            if next_token == "on":
+            if lower(next_token) == "on":
                 next(tokens)  # Consume parenthesis
                 distinct_on_tokens = get_tokens_until_closing_parenthesis(tokens)
                 select_distinct_on = ExpressionListParser.parse(
@@ -83,15 +87,17 @@ class SelectStatementParser:
         )
         expressions = ExpressionListParser.parse(iter(expression_tokens))
 
-        if next_token == "from":
+        if lower(next_token) == "from":
             expression_tokens, next_token = get_tokens_until_one_of(
-                tokens, ["where", "group", "having", "order", "limit", "offset", ";"]
+                tokens,
+                ["where", "group", "having", "order", "limit", "offset", ";"],
+                keep=[("with", "offset")],
             )
             from_statement = FromStatementParser.parse(iter(expression_tokens))
         else:
             from_statement = None
 
-        if next_token == "where":
+        if lower(next_token) == "where":
             expression_tokens, next_token = get_tokens_until_one_of(
                 tokens, ["group", "having", "order", "limit", "offset", ";"]
             )
@@ -99,9 +105,9 @@ class SelectStatementParser:
         else:
             where_clause = None
 
-        if next_token == "group":
+        if lower(next_token) == "group":
             next_token = next(tokens, None)
-            if not next_token == "by":
+            if not lower(next_token) == "by":
                 raise ParsingError("Missing BY after GROUP")
             expression_tokens, next_token = get_tokens_until_one_of(
                 tokens, ["having", "order", "limit", "offset", ";"]
@@ -110,7 +116,7 @@ class SelectStatementParser:
         else:
             group_by_clause = None
 
-        if next_token == "having":
+        if lower(next_token) == "having":
             expression_tokens, next_token = get_tokens_until_one_of(
                 tokens, ["order", "limit", "offset", ";"]
             )
@@ -118,9 +124,9 @@ class SelectStatementParser:
         else:
             having_clause = None
 
-        if next_token == "order":
+        if lower(next_token) == "order":
             next_token = next(tokens, None)
-            if not next_token == "by":
+            if not lower(next_token) == "by":
                 raise ParsingError("Missing BY after ORDER")
             expression_tokens, next_token = get_tokens_until_one_of(
                 tokens, ["limit", "offset", ";"]
@@ -129,7 +135,7 @@ class SelectStatementParser:
         else:
             order_by_clause = None
 
-        if next_token == "limit":
+        if lower(next_token) == "limit":
             expression_tokens, next_token = get_tokens_until_one_of(
                 tokens, ["offset", ";"]
             )
@@ -137,7 +143,7 @@ class SelectStatementParser:
         else:
             limit_clause = None
 
-        if next_token == "offset":
+        if lower(next_token) == "offset":
             expression_tokens, next_token = get_tokens_until_one_of(tokens, [";"])
             offset_clause = OffsetClauseParser.parse(iter(expression_tokens))
         else:
@@ -174,22 +180,18 @@ class FromStatementParser:
         elif next_token == "'" or next_token == '"' or next_token == "`":
             expression = Table(StringParser.parse(tokens, next_token))
         else:
-            if next_token == "unnest":
-                next_next_token = next(tokens)
-                assert next_next_token == "("
-                argument_tokens = get_tokens_until_closing_parenthesis(tokens)
-                arguments = ExpressionListParser.parse(iter(argument_tokens))
-                expression = FunctionCall(next_token, *arguments)
+            if lower(next_token) == "unnest":
+                expression = UnnestParser.parse(tokens)
             else:
                 expression = Table(next_token)
 
         next_token = next(tokens, None)
         if (
             next_token is not None
-            and next_token not in Join.VALUES
-            and next_token not in CombinedQueries.SET_OPERATORS
+            and lower(next_token) not in Join.VALUES
+            and lower(next_token) not in CombinedQueries.SET_OPERATORS
         ):
-            if next_token == "as":
+            if lower(next_token) == "as":
                 with_as = True
                 alias = next(tokens)
             else:
@@ -198,7 +200,7 @@ class FromStatementParser:
             expression = Alias(expression, alias, with_as)
             next_token = next(tokens, None)
 
-        while next_token in CombinedQueries.SET_OPERATORS:
+        while lower(next_token) in CombinedQueries.SET_OPERATORS:
             left_expr = expression
             expression_tokens, next_token = get_tokens_until_not_in(
                 tokens, CombinedQueries.SET_OPERATORS, first_token=next_token
@@ -209,7 +211,7 @@ class FromStatementParser:
                 CombinedQueries.SET_OPERATORS,
                 first_token=next_token,
             )
-            if expression_tokens[0] == "select":
+            if lower(expression_tokens[0]) == "select":
                 right_expr = SelectStatementParser.parse(iter(expression_tokens[1:]))
             else:
                 right_expr = FromStatementParser.parse(iter(expression_tokens))
@@ -217,9 +219,9 @@ class FromStatementParser:
 
             if (
                 next_token is not None
-                and next_token not in CombinedQueries.SET_OPERATORS
+                and lower(next_token) not in CombinedQueries.SET_OPERATORS
             ):
-                if next_token == "as":
+                if lower(next_token) == "as":
                     with_as = True
                     alias = next(tokens)
                 else:
@@ -228,7 +230,7 @@ class FromStatementParser:
                 expression = Alias(expression, alias, with_as)
                 next_token = next(tokens, None)
 
-        while next_token in Join.VALUES:
+        while lower(next_token) in Join.VALUES:
             left_expr = expression
             expression_tokens, next_token = get_tokens_until_not_in(
                 tokens, Join.VALUES, first_token=next_token
@@ -245,19 +247,19 @@ class FromStatementParser:
                 tokens,
                 Join.VALUES,
             )
-            if on_or_using == "on":
+            if lower(on_or_using) == "on":
                 expression = ExpressionParser.parse(iter(expression_tokens))
                 on = OnClause(expression)
-            elif on_or_using == "using":
+            elif lower(on_or_using) == "using":
                 expressions = ExpressionParser.parse(iter(expression_tokens))
                 using = UsingClause(expressions)
-            else:
+            elif join_type not in ("CROSS JOIN", ","):
                 raise ParsingError("Missing ON or USING for join")
 
             expression = Join(join_type, left_expr, right_expr, on=on, using=using)
 
-            if next_token is not None and next_token not in Join.VALUES:
-                if next_token == "as":
+            if next_token is not None and lower(next_token) not in Join.VALUES:
+                if lower(next_token) == "as":
                     with_as = True
                     alias = next(tokens)
                 else:
@@ -276,6 +278,60 @@ class JoinTypeParser:
         return " ".join(tokens).upper()
 
 
+class UnnestParser:
+    @staticmethod
+    def parse(tokens):
+        next_token = next(tokens)
+        assert next_token == "("
+        argument_tokens = get_tokens_until_closing_parenthesis(tokens)
+        arguments = ExpressionListParser.parse(iter(argument_tokens))
+        expression = FunctionCall("unnest", *arguments)
+
+        next_token = next(tokens, None)
+        if (
+            next_token is not None
+            and lower(next_token) not in Join.VALUES
+            and lower(next_token) not in CombinedQueries.SET_OPERATORS
+            and lower(next_token) != "with"
+        ):
+            if lower(next_token) == "as":
+                with_as = True
+                alias = next(tokens)
+            else:
+                with_as = False
+                alias = next_token
+            expression = Alias(expression, alias, with_as)
+            next_token = next(tokens, None)
+
+        with_offset = False
+        with_offset_as = False
+        offset_alias = None
+
+        if lower(next_token) == "with":
+            next_token = next(tokens)
+            assert next_token == "offset"
+            with_offset = True
+
+            next_token = next(tokens, None)
+            if (
+                next_token is not None
+                and lower(next_token) not in Join.VALUES
+                and lower(next_token) not in CombinedQueries.SET_OPERATORS
+            ):
+                if lower(next_token) == "as":
+                    with_offset_as = True
+                    offset_alias = next(tokens)
+                else:
+                    offset_alias = next_token
+
+        return Unnest(
+            expression,
+            with_offset=with_offset,
+            with_offset_as=with_offset_as,
+            offset_alias=offset_alias,
+        )
+
+
 class SetOperatorTypeParser:
     @staticmethod
     def parse(tokens):
@@ -293,7 +349,7 @@ class GroupByParser:
     @staticmethod
     def parse(tokens):
         next_token = next(tokens)
-        if next_token == "rollup":
+        if lower(next_token) == "rollup":
             rollup = True
             next_token = None
         else:
@@ -333,10 +389,10 @@ class OrderByItemParser:
     def parse(tokens):
         expression_tokens, next_token = get_tokens_until_one_of(tokens, ["asc", "desc"])
         expression = ExpressionParser.parse(iter(expression_tokens))
-        if next_token == "asc":
+        if lower(next_token) == "asc":
             has_asc = True
             has_desc = False
-        elif next_token == "desc":
+        elif lower(next_token) == "desc":
             has_asc = False
             has_desc = True
         else:
@@ -350,7 +406,7 @@ class LimitClauseParser:
     @staticmethod
     def parse(tokens):
         next_token = next(tokens)
-        if next_token == "all":
+        if lower(next_token) == "all":
             limit_all = True
             expression = None
         else:
@@ -371,7 +427,7 @@ class OffsetClauseParser:
 
 class ExpressionListParser:
     @staticmethod
-    def parse(tokens):
+    def parse(tokens, can_be_type=False):
         next_token = next(tokens, None)
 
         expressions = []
@@ -385,7 +441,11 @@ class ExpressionListParser:
                 count_parenthesis -= 1
             next_token = next(tokens, None)
             if next_token is None or (next_token == "," and count_parenthesis == 0):
-                expressions.append(ExpressionParser.parse(iter(expression_tokens)))
+                expressions.append(
+                    ExpressionParser.parse(
+                        iter(expression_tokens), can_be_type=can_be_type
+                    )
+                )
                 expression_tokens = []
                 next_token = next(tokens, None)
         return expressions
@@ -393,7 +453,7 @@ class ExpressionListParser:
 
 class ExpressionParser:
     @staticmethod
-    def parse(tokens, is_right_hand=False):
+    def parse(tokens, is_right_hand=False, can_be_type=False):
         main_token = next(tokens)
         next_token = None
 
@@ -401,11 +461,11 @@ class ExpressionParser:
             expression = StringParser.parse(tokens, main_token)
         elif main_token.isdigit():
             expression = Integer(main_token)
-        elif main_token in Boolean.BOOLEAN_VALUES:
+        elif lower(main_token) in Boolean.BOOLEAN_VALUES:
             expression = Boolean(main_token)
-        elif main_token in Null.VALUES:
+        elif lower(main_token) in Null.VALUES:
             expression = Null()
-        elif main_token == Negation.PREDICATE:
+        elif lower(main_token) == Negation.PREDICATE:
             rest_expression, next_token = ExpressionParser.parse(
                 tokens, is_right_hand=True
             )
@@ -414,12 +474,12 @@ class ExpressionParser:
             argument_tokens = get_tokens_until_closing_parenthesis(tokens)
             arguments = ExpressionListParser.parse(iter(argument_tokens))
             expression = Parenthesis(*arguments)
-        elif main_token == "case":
+        elif lower(main_token) == "case":
             argument_tokens, next_token = get_tokens_until_one_of(tokens, ["end"])
-            assert next_token == "end"
+            assert lower(next_token) == "end"
             next_token = next(tokens, None)
             expression = CaseParser.parse(iter(argument_tokens))
-        elif main_token == "select":
+        elif lower(main_token) == "select":
             argument_tokens, next_token = get_tokens_until_one_of(tokens, [])
             next_token = next(tokens, None)
             expression = SelectStatementParser.parse(iter(argument_tokens))
@@ -432,19 +492,33 @@ class ExpressionParser:
         # Expressions that need the next_token to be read
         if expression is None:
             if next_token is not None and next_token == "(":
-                argument_tokens = get_tokens_until_closing_parenthesis(tokens)
-                arguments = ExpressionListParser.parse(iter(argument_tokens))
-                expression = FunctionCall(main_token, *arguments)
+                if lower(main_token) == "cast":
+                    column_tokens, next_token = get_tokens_until_one_of(
+                        tokens, stop_words=["as"]
+                    )
+                    column, _ = ExpressionParser.parse(
+                        iter(column_tokens), is_right_hand=True
+                    )
+                    assert lower(next_token) == "as", next_token
+                    next_token = next(tokens)
+                    cast_type = Type(next_token)
+                    expression = CastFunctionCall(column, cast_type)
+                else:
+                    argument_tokens = get_tokens_until_closing_parenthesis(tokens)
+                    arguments = ExpressionListParser.parse(
+                        iter(argument_tokens), can_be_type=True
+                    )
+                    expression = FunctionCall(main_token, *arguments)
                 next_token = next(tokens, None)
             elif (
                 next_token is not None
-                and main_token in DatePartExtraction.PARTS
-                and next_token == "from"
+                and lower(main_token) in DatePartExtraction.PARTS
+                and lower(next_token) == "from"
             ):
                 rest_expression = ExpressionParser.parse(tokens)
                 expression = DatePartExtraction(main_token, rest_expression)
                 next_token = next(tokens, None)
-            elif main_token in Type.VALUES:
+            elif lower(main_token) in Type.VALUES and can_be_type:
                 expression = Type(main_token)
             elif next_token is not None and next_token == "[":
                 argument_tokens, next_token = get_tokens_until_one_of(
@@ -459,9 +533,9 @@ class ExpressionParser:
                 expression = Integer(-int(next_token))
                 next_token = next(tokens, None)
             elif (
-                main_token in String.PREFIXES
+                lower(main_token) in String.PREFIXES
                 and next_token is not None
-                and next_token in String.QUOTES
+                and lower(next_token) in String.QUOTES
             ):
                 expression = StringParser.parse(
                     tokens, start_quote=next_token, prefix=main_token
@@ -469,16 +543,16 @@ class ExpressionParser:
             else:
                 expression = Column(main_token)
 
-        if next_token == "over":
+        if lower(next_token) == "over":
             opening_parenthesis = next(tokens, None)
             if opening_parenthesis != "(":
                 raise ParsingError("expected '('")
 
             argument_tokens = iter(get_tokens_until_closing_parenthesis(tokens))
             argument_next_token = next(argument_tokens, None)
-            if argument_next_token == "partition":
+            if lower(argument_next_token) == "partition":
                 argument_next_token = next(argument_tokens, None)
-                if not argument_next_token or argument_next_token != "by":
+                if not argument_next_token or lower(argument_next_token) != "by":
                     raise ParsingError("Missing BY after PARTITION")
                 expression_tokens, argument_next_token = get_tokens_until_one_of(
                     argument_tokens, ["order", "rows", "range"]
@@ -487,9 +561,9 @@ class ExpressionParser:
             else:
                 partition_by = None
 
-            if argument_next_token == "order":
+            if lower(argument_next_token) == "order":
                 argument_next_token = next(argument_tokens, None)
-                if not argument_next_token or argument_next_token != "by":
+                if not argument_next_token or lower(argument_next_token) != "by":
                     raise ParsingError("Missing BY after ORDER")
                 expression_tokens, argument_next_token = get_tokens_until_one_of(
                     argument_tokens, ["rows", "range"]
@@ -498,7 +572,7 @@ class ExpressionParser:
             else:
                 order_by = None
 
-            if argument_next_token in ("rows", "range"):
+            if lower(argument_next_token) in ("rows", "range"):
                 rows_range = argument_next_token
                 expression_tokens, _ = get_tokens_until_one_of(argument_tokens, [])
                 frame_clause = WindowFrameClause(
@@ -524,11 +598,11 @@ class ExpressionParser:
         if is_right_hand:
             return expression, next_token
 
-        if next_token in Condition.PREDICATES:
+        if lower(next_token) in Condition.PREDICATES:
             symbol = next_token
-            if next_token == "is":
+            if lower(next_token) == "is":
                 next_next_token = next(tokens)
-                if next_next_token == "not":
+                if lower(next_next_token) == "not":
                     symbol = "is not"
                 else:
                     tokens, _ = get_tokens_until_one_of(
@@ -538,12 +612,12 @@ class ExpressionParser:
 
             right_hand, next_token = ExpressionParser.parse(tokens, is_right_hand=True)
             expression = Condition(expression, symbol, right_hand)
-        elif next_token == "between":
+        elif lower(next_token) == "between":
             symbol = next_token
             right_hand_left, next_token = ExpressionParser.parse(
                 tokens, is_right_hand=True
             )
-            if next_token != "and":
+            if lower(next_token) != "and":
                 raise ParsingError("expected AND")
             right_hand_right, next_token = ExpressionParser.parse(
                 tokens, is_right_hand=True
@@ -554,15 +628,19 @@ class ExpressionParser:
                 right_hand_right,
             )
             expression = Condition(expression, symbol, right_hand)
+        elif next_token in BitwiseOperation.OPERATORS:
+            operator = next_token
+            right_hand, next_token = ExpressionParser.parse(tokens, is_right_hand=True)
+            expression = BitwiseOperation(expression, operator, right_hand)
 
-        if next_token in BooleanCondition.PREDICATES:
+        if lower(next_token) in BooleanCondition.PREDICATES:
             left_hand = expression
             symbol = next_token
             right_hand = ExpressionParser.parse(tokens)
             expression = BooleanCondition(symbol, left_hand, right_hand)
             next_token = next(tokens, None)
 
-        if next_token == "except":
+        if lower(next_token) == "except":
             opening_parenthesis = next(tokens, None)
             if opening_parenthesis != "(":
                 raise ParsingError("expected '('")
@@ -579,7 +657,7 @@ class ExpressionParser:
             and next_token != "`"
             and next_token != ";"
         ):
-            if next_token == "as":
+            if lower(next_token) == "as":
                 with_as = True
                 alias, _ = ExpressionParser.parse(tokens, is_right_hand=True)
             else:
@@ -609,7 +687,7 @@ class CaseParser:
     @staticmethod
     def parse(tokens):
         next_token = next(tokens)
-        if next_token == "when":
+        if lower(next_token) == "when":
             expression = None
         else:
             expressions_tokens, next_token = get_tokens_until_one_of(
@@ -620,7 +698,7 @@ class CaseParser:
         when_then = []
         else_expression = None
         while next_token:
-            if next_token == "when":
+            if lower(next_token) == "when":
                 expressions_tokens, _ = get_tokens_until_one_of(tokens, ["then"])
                 when_expression = ExpressionParser.parse(iter(expressions_tokens))
                 expressions_tokens, next_token = get_tokens_until_one_of(
@@ -628,7 +706,7 @@ class CaseParser:
                 )
                 then_expression = ExpressionParser.parse(iter(expressions_tokens))
                 when_then.append((when_expression, then_expression))
-            elif next_token == "else":
+            elif lower(next_token) == "else":
                 expressions_tokens, next_token = get_tokens_until_one_of(tokens, [])
                 else_expression = ExpressionParser.parse(iter(expressions_tokens))
 
