@@ -1,5 +1,7 @@
 from typing import Any
 
+from sqlvalidator.grammar.tokeniser import lower
+
 DEFAULT_LINE_LENGTH = 88
 
 
@@ -81,7 +83,7 @@ class SelectStatement:
             statement_str += "\nFROM {}".format(from_str)
 
         if self.where_clause:
-            statement_str += "\nWHERE {}".format(transform(self.where_clause))
+            statement_str += "\nWHERE{}".format(transform(self.where_clause))
 
         if self.group_by_clause:
             statement_str += "\nGROUP BY{}".format(transform(self.group_by_clause))
@@ -223,28 +225,12 @@ class Expression:
 
 class WhereClause(Expression):
     def transform(self):
-        """
-        overriden_value = bool(value)
-        value = value or self.value
-
-        is_parenthesis = isinstance(value, Parenthesis)
-        if is_parenthesis:
-            value = value.args[0]
-
-        if isinstance(value, BooleanCondition):
-            join_str = "\n{} ".format(value.type.upper())
-            where_str = "\n" + join_str.join(self.transform(a) for a in value.args)
-            where_str = where_str.replace("\n", "\n ")
-        elif overriden_value:
-            where_str = str(value)
-        else:
-            where_str = " {}".format(value)
-
-        if is_parenthesis:
-            where_str = " ({}\n)".format(where_str)
-        return where_str
-        """
-        return str(self.value)
+        transformed_value = transform(self.value)
+        if isinstance(self.value, Parenthesis) and "\n" in transformed_value:
+            return " (\n " + transform(self.value.args[0]) + "\n)"
+        if "\n" in transformed_value:
+            return "\n " + transformed_value
+        return " " + transformed_value
 
     def validate(self, known_fields):
         errors = super().validate(known_fields)
@@ -309,6 +295,9 @@ class GroupByClause(Expression):
 
 
 class HavingClause(Expression):
+    def __str__(self):
+        return transform(self.value)
+
     def validate(self, known_fields):
         errors = super().validate(known_fields)
         errors += self.value.validate(known_fields)
@@ -476,6 +465,18 @@ class FunctionCall(Expression):
         )
 
 
+class CastFunctionCall(FunctionCall):
+    def __init__(self, column, cast_type):
+        super().__init__("cast", column, "AS", cast_type)
+
+    def __str__(self):
+        return "{}({} AS {})".format(
+            self.function_name.upper(),
+            transform(self.args[0]),
+            transform(self.args[2]),
+        )
+
+
 class AnalyticsClause(Expression):
     def __init__(self, function, partition_by, order_by, frame_clause):
         self.function = function
@@ -593,7 +594,7 @@ class Column(Expression):
 
 
 class Type(Expression):
-    VALUES = ("int", "float", "day", "month", "timestamp", "int64", "string")
+    VALUES = ("int", "float", "day", "month", "timestamp", "int64", "string", "date")
 
     def __str__(self):
         return self.value.upper()
@@ -696,6 +697,7 @@ class Boolean(Expression):
     BOOLEAN_VALUES = TRUE_VALUES + FALSE_VALUES
 
     def __init__(self, value):
+        value = lower(value)
         assert value in self.BOOLEAN_VALUES
         if value in self.TRUE_VALUES:
             value = True
@@ -832,12 +834,59 @@ class Addition(ArithmaticOperator):
         super().__init__("+", *args)
 
 
+class BitwiseOperation(Expression):
+    OPERATORS = ("&", "|", "^")
+
+    def __init__(self, expression, predicate, right_hand):
+        super().__init__(expression)
+        self.predicate = predicate
+        self.right_hand = right_hand
+
+    def __str__(self):
+        return "{} {} {}".format(
+            transform(self.value), self.predicate, transform(self.right_hand)
+        )
+
+    def __repr__(self):
+        return "<BitwiseOperation: {!r} {} {!r}>".format(
+            self.value, self.predicate, self.right_hand
+        )
+
+
 class Table(Expression):
     pass
 
 
+class Unnest(Expression):
+    def __init__(self, unnest_expression, with_offset, with_offset_as, offset_alias):
+        # unnest_expression: can be functiion call or alias of function call
+        super().__init__(unnest_expression)
+        self.with_offset = with_offset
+        self.with_offset_as = with_offset_as
+        self.offset_alias = offset_alias
+
+    def __str__(self):
+        unnest_str = transform(self.value)
+        if self.with_offset:
+            unnest_str += " WITH OFFSET"
+            if self.offset_alias:
+                if self.with_offset_as:
+                    unnest_str += " AS"
+                unnest_str += f" {self.offset_alias}"
+        return unnest_str
+
+    def __eq__(self, other):
+        return (
+            (isinstance(other, type(self)) or isinstance(self, type(other)))
+            and self.value == other.value
+            and self.with_offset == other.with_offset
+            and self.with_offset_as == other.with_offset_as
+            and self.offset_alias == other.offset_alias
+        )
+
+
 class Join(Expression):
-    VALUES = ("join", "inner", "left", "right", "full", "cross", "outer")
+    VALUES = ("join", "inner", "left", "right", "full", "cross", "outer", ",")
 
     def __init__(self, join_type, left_from, right_from, on, using):
         self.join_type = join_type
@@ -912,15 +961,18 @@ class Join(Expression):
                 left_element, " AS" if alias.with_as else "", alias.alias
             )
 
-        join_str = "{}\n{}{}\n".format(
+        join_type_str = (
+            "\n{}".format(self.join_type.upper()) if self.join_type != "," else ","
+        )
+        join_str = "{}{}{}".format(
             left_element,
-            self.join_type.upper(),
+            join_type_str,
             right_element,
         )
         if self.on:
-            join_str += "ON {}".format(transform(self.on))
+            join_str += "\nON {}".format(transform(self.on))
         elif self.using:
-            join_str += "USING {}".format(transform(self.using))
+            join_str += "\nUSING {}".format(transform(self.using))
         return join_str
 
     def __eq__(self, other):
@@ -990,7 +1042,8 @@ class CombinedQueries(Expression):
 
 
 class OnClause(Expression):
-    pass
+    def __str__(self):
+        return transform(self.value)
 
 
 class UsingClause(Expression):
@@ -1006,7 +1059,9 @@ class Condition(Expression):
         self.right_hand = right_hand
 
     def __str__(self):
-        return "{} {} {}".format(self.value, self.predicate.upper(), self.right_hand)
+        return "{} {} {}".format(
+            transform(self.value), self.predicate.upper(), transform(self.right_hand)
+        )
 
     def __repr__(self):
         return "<Condition: {!r} {} {!r}>".format(
@@ -1028,13 +1083,34 @@ class BooleanCondition(Expression):
     PREDICATES = ("and", "or")
 
     def __init__(self, type, *args):
-        assert type in ("and", "or")
+        assert type.lower() in ("and", "or")
         self.type = type
         self.args = args
 
-    def __str__(self):
+    def transform(self, with_newline=False):
         join_str = " {} ".format(self.type.upper())
-        return join_str.join(map(str, self.args))
+        transformed_condition = join_str.join(map(transform, self.args))
+        if len(transformed_condition) < DEFAULT_LINE_LENGTH and with_newline is False:
+            return transformed_condition
+
+        join_str = "\n {} ".format(self.type.upper())
+        transformed_args = []
+        for a in self.args:
+            if isinstance(a, BooleanCondition):
+                transformed_a = a.transform(with_newline=True)
+            elif isinstance(a, Parenthesis):
+                transformed_a = transform(a.args[0])
+                if "\n" in transformed_a:
+                    transformed_a = (
+                        "(\n  " + transformed_a.replace("\n", "\n ") + "\n )"
+                    )
+                else:
+                    transformed_a = f"({transformed_a})"
+            else:
+                transformed_a = transform(a)
+            transformed_args.append(transformed_a)
+
+        return join_str.join(transformed_args)
 
     def __repr__(self):
         return "<BooleanCondition {}: {}>".format(
